#define NORMALMAP
struct Light {
    vec3 direction;
    vec3 intensity;
    float attenuation;
};

varying vec2 vTexcoord;
varying vec3 vNormal;
varying vec3 vWorldPosition;
varying vec3 vTangent;

uniform sampler2D diffuseMap;
uniform sampler2D normalMap;
uniform sampler2D mraMap;

uniform vec3 cameraPosition;

uniform float fresnelIntense;

uniform float specularIntensity;
uniform float wrapped;
uniform float opacity;

#define MEDIUMP_FLT_MAX    65504.0
#define saturateMediump(x) min(x, MEDIUMP_FLT_MAX)
const float PI = 3.14159265359;

struct BRDFParam {
    float metallic  ;
    float roughness ;
    float ao        ;
};

uniform vec4 colorParam;

#ifdef COLORMASKMAP
uniform sampler2D colorMaskMap;
#endif
 
vec3 colorComposite(vec3 compositeColor, vec3 inputColor) { 
    return inputColor;
}
 
#define NR_DIR_LIGHTS 2
uniform sampler2D brdfLut;
uniform samplerCube prefilterMap;
uniform samplerCube irradianceMap;
uniform float IBLIntensity;
uniform float IBLAngle;

vec3 FresnelSchlickRoughness(float cosTheta, vec3 F0, float roughness) {
    return F0 + (max(vec3(1.0 - roughness), F0) - F0) * pow(max(1.0 - cosTheta, 0.0), 5.0);
}

vec3 getIBLColor(vec3 N, vec3 V, vec3 albedo, BRDFParam p) {
    float angle = IBLAngle * 3.14;
    mat4 rotateMat = mat4(cos(angle), 0.0, sin(angle), 0.0, 0.0, 1.0, 0.0, 0.0, -sin(angle), 0.0, cos(angle), 0.0, 0.0, 0.0, 0.0, 1.0);
    vec3 view = V;        
    
    float roughness =  sqrt(p.roughness);
    roughness = clamp(roughness, 0.0, 1.0);

    vec3 F0 = vec3(fresnelIntense);
    F0 = mix(F0, albedo.rgb, p.metallic);
    vec3 F = FresnelSchlickRoughness(max(dot(N, view), 0.0), F0, roughness);

    vec3 kS = F;
    vec3 kD = 1.0 - kS;
    kD *= 1.0 - p.metallic;     
    vec3 R = reflect(-view, N);   
    R = (rotateMat * vec4(R, 1.0)).rgb;

   vec3 normal = (rotateMat * vec4(N, 1.0)).rgb;

    vec3 irradiance = textureCube(irradianceMap, normal).rgb;
    vec3 diffuse    = irradiance * albedo;

    const float MAX_REFLECTION_LOD = 7.0;
    vec3 prefilteredColor = textureCube(prefilterMap, R,  roughness * MAX_REFLECTION_LOD).rgb;
    float preBRDFX = clamp(dot(N, view), 0.01, 0.99);
    float preBRDFY = clamp(1.0 - roughness, 0.01, 0.99);
    vec2 envBRDF  = texture2D(brdfLut, vec2(preBRDFX, preBRDFY)).rg;
    vec3 specular = prefilteredColor * (F * envBRDF.x + envBRDF.y);

    vec3 ambient = (kD * diffuse + specular * IBLIntensity) * p.ao;
    return ambient;
}
#define IBL
//data struct
#define MAX_LIGHT_NUM 8
struct DirLight {
    vec3 position;
    vec3 direction;
    vec3 intensity;
    vec3 intenseColor;
};

struct PointLight {
    vec3 position;
    vec3 intensity;
};

struct SpotLight {
    vec3 position;
    vec3 direction;
    vec3 intensity;
    float cutOff;
    float outerCutOff;
};

//global param
int lightNum = 0;
Light lights[MAX_LIGHT_NUM];

#ifdef NR_DIR_LIGHTS
	uniform DirLight dirLights[NR_DIR_LIGHTS];
#endif

#ifdef NR_POINT_LIGHTS
	uniform PointLight pointLights[NR_POINT_LIGHTS];
#endif

#ifdef NR_SPOT_LIGHTS
	uniform SpotLight spotLights[NR_SPOT_LIGHTS];
#endif

//struct light was definated in main BRDF
Light parseDirLight(vec3 worldPosition, DirLight dirLight) {
    Light ans;
    ans.direction = dirLight.direction;
    ans.intensity = dirLight.intensity * dirLight.intenseColor;
    float attenuation;
    //vWorldPosition comes from main brdf
    float distance = length(dirLight.position - worldPosition);
    attenuation = 1.0 / 5000.0;// / (1.0/*constant*/ + 0.09/*linear*/ * distance + 0.032/*quadratic*/ * (distance * distance));
    ans.attenuation = attenuation;
    return ans;
}

Light parsePointLight(vec3 worldPosition, PointLight pointLight) {
    Light ans;
    ans.direction = normalize(pointLight.position - worldPosition);
    ans.intensity = pointLight.intensity;
    float attenuation;
    
    float distance = length(pointLight.position - worldPosition);
    attenuation = 1.0 / 5000.0;// / (1.0/*constant*/ + 0.09/*linear*/ * distance + 0.032/*quadratic*/ * (distance * distance));
    ans.attenuation = attenuation;
    return ans;
}

Light parseSpotLight(vec3 worldPosition, SpotLight spotLight) {
    Light ans;
    ans.direction = spotLight.direction;

    vec3 lightDir = normalize(spotLight.position - worldPosition);
    float theta     = dot(lightDir, normalize(-spotLight.direction));
    float epsilon   = spotLight.cutOff - spotLight.outerCutOff;
    float intensity = clamp((theta - spotLight.outerCutOff) / epsilon, 0.0, 1.0);
    ans.intensity = intensity * spotLight.intensity;

    float attenuation;

    float distance = length(spotLight.position - vWorldPosition);
    attenuation = 1.0 / 5000.0 ;// / (1.0/*constant*/ + 0.09/*linear*/ * distance + 0.032/*quadratic*/ * (distance * distance));
    ans.attenuation = attenuation;
    return ans;
}


void getLights(vec3 worldPosition) {
    int tmpNum = 0;
#ifdef NR_DIR_LIGHTS
    for (int i = 0; i < NR_DIR_LIGHTS; i++) {
        lights[tmpNum] = parseDirLight(worldPosition, dirLights[i]);
        tmpNum++;
    }
    lightNum += NR_DIR_LIGHTS;
#endif
#ifdef NR_POINT_LIGHTS
    for (int i = 0; i < NR_POINT_LIGHTS; i++) {
        lights[tmpNum] = parsePointLight(worldPosition, pointLights[i]);
        tmpNum++;
    }
    lightNum += NR_POINT_LIGHTS;
#endif
#ifdef NR_SPOT_LIGHTS
    for (int i = 0; i < NR_SPOT_LIGHTS; i++) {
        lights[tmpNum] = parseSpotLight(worldPosition, spotLights[i]);
        tmpNum++;
    }
    lightNum += NR_SPOT_LIGHTS;
#endif
}

vec3 getNormalFromMap(){
    vec3 tangentNormal = texture2D(normalMap, vTexcoord).rgb * 2.0 - vec3(1.0);

    vec3 N = normalize(vNormal);
    vec3 T = normalize(vTangent);
    vec3 B = normalize(cross(N,T));
    mat3 TBN = mat3(T,B,N);
    return normalize(TBN * normalize(tangentNormal));
}

//cosTheta : n dot v
vec3 fresnelSchlick(float cosTheta, vec3 F0){
    return F0 + (1.0 - F0) * pow(max(1.0 - cosTheta, 0.0), 5.0);
}

float V_SmithGGXCorrelatedFast(float NoV, float NoL, float roughness) {
    float a = roughness * roughness;
    float GGXV = NoL * (NoV * (1.0 - a) + a);
    float GGXL = NoV * (NoL * (1.0 - a) + a);
    return 0.5 / (GGXV + GGXL);
}

float D_GGX(const vec3 n, const vec3 h, float roughness) {
    vec3 NxH = cross(n, h);
    float NoH = dot(n,h);
    float a = NoH * roughness;
    float k = roughness / (dot(NxH, NxH) + a * a);
    float d = k * k * (1.0 / PI);
    return saturateMediump(d);
}

const vec2 invAtan = vec2(0.1591, 0.3183);
vec2 SampleSphericalMap(vec3 v)
{
    vec2 uv = vec2(atan(v.z, v.x), asin(v.y));
    uv *= invAtan;
    uv += 0.5;
    return uv;
}

vec3 BRDF(Light light, BRDFParam p, vec3 N, vec3 V, vec3 albedo) {
    
    vec3 F0 = vec3(fresnelIntense);
    F0 = mix(F0, albedo.rgb, p.metallic);
    
    vec3 L = -light.direction;
    vec3 H = normalize(V + L);
    float NoV = abs(dot(N, V)) + 1e-5;
    float NoL = clamp(dot(N, L), 0.0, 1.0);
    float LoH = clamp(dot(H, L), 0.0, 1.0);
//base brdf
//pre calculate of Standard brdf
    vec3 radiance = light.attenuation * light.intensity;
    vec3 F    = fresnelSchlick(LoH, F0); 

    vec3 energyCompensation = vec3(1.0);// + F0 * (1.0 / dfg.y - 1.0);

#ifdef ANISOTROPIC
    float NDFG = getAnitropicMicroNormal(N, L, H, V, p.roughness);
    vec3 specular = NDFG * F * specularIntensity * energyCompensation;
#else
    float NDF = D_GGX(N, H, p.roughness);
    float G   = V_SmithGGXCorrelatedFast(NoV, NoL, p.roughness);
    vec3 nominator = NDF * G * F;
    vec3 specular = nominator * specularIntensity * energyCompensation;
#endif
    
    vec3 kS = F;
    vec3 kD = vec3(1.0) - kS;
    kD *= 1.0 - p.metallic;
    float NdotL = max(dot(N, L), 0.0);
    vec3 diffuseColor = kD * albedo.rgb / PI * (NdotL + wrapped) / ((1.0 + wrapped) * (1.0 + wrapped));
#ifdef CLEARCOAT
    float VoH = clamp(dot(V, H), 0.0, 1.0);
    float Frc = getClearCoatPower(N, H, V, L);
    vec3 color = (specular * (1.0 - Frc)* (1.0 - Frc) + Frc) * NdotL + diffuseColor;
#else
    vec3 color = specular * NdotL + diffuseColor;
#endif

    color = color * radiance;

#ifdef SUBSURFACE
    color += getSubsurfaceColor(kD, radiance, V, -light.direction, N, albedo);
#endif

    return color;
}

uniform float metallicParam;
uniform float roughnessParam;
void main()
{
    vec4 mapColor = texture2D(diffuseMap, vTexcoord).rgba;
    vec3 albedo = mapColor.rgb; 
    if (colorParam.w > 0.0) {
        float colorMask = 1.0;
#ifdef COLORMASKMAP
        colorMask = texture2D(colorMaskMap, vTexcoord).r;
#endif 
        albedo = colorMask * colorComposite(colorParam.rgb, albedo) + (1.0 - colorMask) * albedo;
    }
    albedo = pow(albedo, vec3(2.2)); 
    getLights(vWorldPosition);
    BRDFParam p;
    BRDFParam IBLBrdfParam;
    vec3 mraColor   = texture2D(mraMap,vTexcoord).rgb;
#ifdef METALLICPARAM
    p.metallic  = clamp(metallicParam, 0.01, 1.0);
#else
    p.metallic  = clamp(mraColor.r * metallicParam, 0.01, 1.0);
#endif

    IBLBrdfParam.metallic = p.metallic;
#ifdef ROUGHNESSPARAM
    p.roughness =  roughnessParam;
#else
    if (roughnessParam < 1.0) {
        p.roughness = (mraColor.g) * roughnessParam;
    } else {
        p.roughness = (mraColor.g) + (roughnessParam - 1.0) / 2.0 * (1.0 - mraColor.g);
    }
#endif
    IBLBrdfParam.roughness = p.roughness;
    p.roughness = clamp(p.roughness * p.roughness, 0.01, 1.0);

    p.ao        = mraColor.b;
    IBLBrdfParam.ao = p.ao;

#ifdef NORMALMAP 
    vec3 N = getNormalFromMap();
#else
    vec3 N = normalize(vNormal);
#endif
    if (!gl_FrontFacing) {
        N = -N;
    }
    vec3 V = normalize(cameraPosition - vWorldPosition);
    vec3 R = reflect(-V, N);
    
    vec3 Lo = vec3(0.0);
//base brdf
//pre calculate of Standard brdf
    for(int i = 0; i < lightNum; i++) {
        Lo += BRDF(lights[i], p, N, V, albedo);
#ifdef RIMLIGHT
        Lo += rimLight(V, -lights[i].direction, N, albedo); 
#endif
    } 

#ifdef IBL
    Lo += getIBLColor(N, V, albedo, IBLBrdfParam); 
#endif

    vec3 color = Lo;//ambient + Lo;
    color = pow(color , vec3(1.0 / 2.2));
    
    gl_FragColor = vec4(color, mapColor.a * opacity);
}
